"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
Object.defineProperty(exports, "__esModule", { value: true });
var electron_1 = require("electron");
var os = require("os");
var fs = require("fs");
var path = require("path");
var url = require("url");
var windowStateKeeper = require("electron-window-state");
var commander = require("commander");
var win;
var isRelease = false;
function createApp() {
    //Fix up empty node command line in bundled mode (crashes commander otherwise)
    if (electron_1.app.isPackaged) {
        process.argv[0] = 'main.js';
        process.argv.unshift('node');
        //Fix PATH on macOS when bundled
        if (os.platform() == "darwin") {
            process.env.PATH = ['/usr/local/bin', process.env.PATH].join(':');
        }
    }
    //Parse command line
    commander
        .option('p, python <path>', 'Path to your python executable')
        .option('r, release', 'Runs electron in release mode')
        .parse(process.argv);
    global["commandLineArgs"] = commander;
    isRelease = commander.release || electron_1.app.isPackaged;
    //Load the previous window state with fallback to defaults
    var mainWindowState = windowStateKeeper({
        defaultWidth: 961,
        defaultHeight: 888
    });
    //Browser Window common options
    var browserOptions = { icon: path.join(__dirname, '../src/assets/icon/png/64x64.png'), title: 'OoT Randomizer GUI', opacity: 1.00, backgroundColor: '#000000', minWidth: 880, minHeight: 680, width: mainWindowState.width, height: mainWindowState.height, x: mainWindowState.x, y: mainWindowState.y, show: false, webPreferences: { nodeIntegration: false, contextIsolation: true, webviewTag: false, preload: path.join(__dirname, 'preload.js') } };
    //Override menu (only need dev tools shortcut)
    var appMenu = new electron_1.Menu();
    var subMenu = new electron_1.Menu();
    subMenu.append(new electron_1.MenuItem({
        label: 'Toggle Developer Tools',
        accelerator: os.platform() === 'darwin' ? 'Alt+Cmd+I' : 'Ctrl+Shift+I',
        click: function () { win.webContents.openDevTools(); }
    }));
    subMenu.append(new electron_1.MenuItem({
        label: 'Refresh GUI',
        accelerator: 'F5',
        click: function () { win.webContents.reload(); }
    }));
    appMenu.append(new electron_1.MenuItem({
        label: 'Electron',
        type: 'submenu',
        role: null,
        accelerator: null,
        submenu: subMenu
    }));
    //macOS specific overrides
    if (os.platform() == "darwin") {
        browserOptions["titleBarStyle"] = 'hiddenInset'; //macOS uses titleBarStyle to maintain a native feel with the finder menu and shortcuts users expect
        //Alter the dock icon on macOS and make it bounce to indicate activity until the browser window is created
        electron_1.app.dock.setIcon(path.join(__dirname, '../src/assets/icon/png/64x64.png'));
        electron_1.app.dock.bounce("critical");
    }
    else {
        browserOptions["frame"] = false; //Hide menu and frame entirely on every other platform since we use our custom title bar
        browserOptions["fullscreen"] = false; //Fullscreen mode is confusing with our custom title bar on Windows/Linux
        browserOptions["fullscreenable"] = false;
    }
    win = new electron_1.BrowserWindow(browserOptions);
    win.setMenu(appMenu);
    //WindowStateKeeper will automatically persist window state changes to file
    mainWindowState.manage(win);
    //Run Content Security Policy
    manageCSP();
    if (!isRelease) {
        win.loadURL("http://localhost:4200/"); //Dev server
    }
    else { //Load release dist
        //Check if it exists first
        var indexPath = path.join(__dirname, '../../dist/ootr-electron-gui/index.html');
        if (!fs.existsSync(indexPath)) {
            console.error("No release found");
            setTimeout(function () {
                electron_1.app.quit();
            }, 0);
            throw Error("No Electron GUI found! Did you compile it properly?");
        }
        win.loadURL(url.format({
            pathname: indexPath,
            protocol: "file:",
            slashes: true
        }));
    }
    win.once('ready-to-show', function () {
        win.show();
        if (!isRelease && !win.isMaximized()) //Open dev tools automatically if dev mode and not maximized
            win.webContents.openDevTools();
    });
    //macOS exclusive, goes to idle state
    win.on("closed", function () {
        win = null;
    });
}
//LISTENERS
electron_1.app.on("ready", createApp);
//macOS exclusive, handles soft re-launches
electron_1.app.on("activate", function () {
    if (win === null) {
        createApp();
    }
});
electron_1.app.on("window-all-closed", function () {
    //Ensures the electron process always shuts down properly if all windows have been closed
    //Don't do this on macOS as users expect to be able to re-launch the app quickly from the dock after all windows get closed
    if (os.platform() != "darwin") {
        setTimeout(function () {
            electron_1.app.quit();
        }, 1000);
    }
});
//Limit navigation to safe URLs and defer unsafe popups to system browser as well as handle page loading errors
electron_1.app.on('web-contents-created', function (event, contents) {
    contents.on('will-navigate', function (event, navigationUrl) {
        var parsedUrl = new URL(navigationUrl);
        //console.log("Navigation attempt to:", parsedUrl.origin);
        //Whitelist for dev server in dev mode
        if (!isRelease) {
            if (parsedUrl.origin === 'http://localhost:4200') {
                return;
            }
        }
        event.preventDefault();
    });
    contents.on('new-window', function (event, navigationUrl) {
        var parsedUrl = new URL(navigationUrl);
        //console.log("New window creation attempt:", parsedUrl.origin);
        //Whitelist for dev server in dev mode
        if (!isRelease) {
            if (parsedUrl.origin === 'http://localhost:4200') {
                return;
            }
        }
        event.preventDefault();
        electron_1.shell.openExternal(navigationUrl);
    });
    contents.on("did-fail-load", function () {
        console.error("GUI load failed");
        setTimeout(function () {
            electron_1.app.quit();
        }, 0);
        if (!isRelease)
            throw Error("Couldn't connect to localhost:4200. Please start the Angular development server first before attempting to boot the Electron GUI!");
        else
            throw Error("The Electron GUI failed to load. Please delete the dist folder and re-compile!");
    });
});
//CSP
function manageCSP() {
    electron_1.session.defaultSession.webRequest.onHeadersReceived(function (details, callback) {
        callback({
            responseHeaders: __assign({}, details.responseHeaders, { 'Content-Security-Policy': ['default-src \'self\' *.gstatic.com; img-src \'self\' data: *; style-src \'self\' \'unsafe-inline\' *.googleapis.com; script-src \'self\' \'unsafe-eval\' \'unsafe-inline\' *.googleapis.com; connect-src \'self\' *.githubusercontent.com ws: localhost:4200*'] })
        });
    });
}
//IPC
electron_1.ipcMain.on('getGeneratorGUISettings', function (event, arg) {
    var pythonRootPath = electron_1.app.isPackaged ? electron_1.app.getAppPath() + "/python/" : electron_1.app.getAppPath() + "/../";
    //Load compiled settings_list.json
    var compiledSettingsMapPath = path.normalize(pythonRootPath + "data/generated/settings_list.json");
    var guiSettings;
    if (fs.existsSync(compiledSettingsMapPath)) {
        guiSettings = JSON.parse(fs.readFileSync(compiledSettingsMapPath, 'utf8'));
    }
    else {
        console.error("No settings_list.json found!");
        setTimeout(function () {
            electron_1.app.quit();
        }, 0);
        throw Error("No settings_list.json found! Please restart the GUI using the Gui.py");
    }
    //Add static presets
    guiSettings.presets = {
        "[New Preset]": { isNewPreset: true },
        "Default / Beginner": { isDefaultPreset: true }
    };
    //Load built in presets
    var builtInPresetsPath = path.normalize(pythonRootPath + "data/presets_default.json");
    if (fs.existsSync(builtInPresetsPath)) {
        var builtInPresets_1 = JSON.parse(fs.readFileSync(builtInPresetsPath, 'utf8'));
        var adjustedBuiltInPresets_1 = {};
        //Tag built in presets appropiately
        Object.keys(builtInPresets_1).forEach(function (presetName) {
            if (!(presetName in guiSettings.presets))
                adjustedBuiltInPresets_1[presetName] = { isProtectedPreset: true, settings: builtInPresets_1[presetName] };
        });
        Object.assign(guiSettings.presets, adjustedBuiltInPresets_1);
    }
    //Load user presets
    var userPresetPath = path.normalize(pythonRootPath + "presets.sav");
    if (fs.existsSync(userPresetPath)) {
        var userPresets_1 = JSON.parse(fs.readFileSync(userPresetPath, 'utf8'));
        var adjustedUserPresets_1 = {};
        //Tag user presets appropiately
        Object.keys(userPresets_1).forEach(function (presetName) {
            if (!(presetName in guiSettings.presets))
                adjustedUserPresets_1[presetName] = { settings: userPresets_1[presetName] };
        });
        Object.assign(guiSettings.presets, adjustedUserPresets_1);
    }
    event.returnValue = JSON.stringify(guiSettings);
});
//# sourceMappingURL=main.js.map